"""Provides certain utilities for the streamlit demo."""

from abc import ABC, abstractmethod
from itertools import product, repeat
from typing import List, Iterable, Tuple, Optional

import matplotlib.pyplot as plt
import numpy as np
import streamlit as st

from tnmf.TransformInvariantNMF import TransformInvariantNMF
from tnmf.utils.signals import generate_pulse_train, generate_block_image

HELP_CHANNEL = \
    '''The **number of channels** each input signal comprises. In contrast to the remaining signal dimensions, which are
    treated shift-invariant (meaning that atom placement is flexible), channels represent the inflexible part of the
    factorization in the sense that each atom always covers all channels.'''


def st_define_nmf_params(default_params: dict, verbose: bool = True) -> dict:
    """
    Defines all necessary NMF parameters via streamlit widgets.

    Parameters
    ----------
    default_params : dict
        Contains the default parameters that are used if the created streamlit checkbox is True.
    verbose : bool
        If True, show detailed information.

    Returns
    -------
    nmf_params : dict
        A dictionary containing the selected NMF parameters.
    """
    st.sidebar.markdown('# TNMF settings')

    # decide if ground truth atom number shall be used
    help_use_n_atoms = \
        '''If selected, the **number of atoms** used by the model is set to the actual number of atoms used
        for signal generation.'''
    use_n_atoms = st.sidebar.checkbox('Use ground truth number of atoms', True, help=help_use_n_atoms)
    if verbose:
        st.sidebar.caption(help_use_n_atoms)

    help_n_atoms = 'The **number of atoms**. To use the ground truth dictionary size, tick the checkbox above.'
    n_atoms = st.sidebar.number_input('# Atoms', value=default_params['n_atoms'], min_value=1,
                                      help=help_n_atoms) if not use_n_atoms else None
    if verbose and not use_n_atoms:
        st.sidebar.caption(help_n_atoms)

    # decide if ground truth atom shape shall be used
    help_use_atom_shape = \
        '''If selected, the **size of the atoms** used by the model is set the actual size of the atoms used
        for signal generation.'''
    use_atom_shape = st.sidebar.checkbox('Use ground truth atom size', True, help=help_use_atom_shape)
    if verbose:
        st.sidebar.caption(help_use_atom_shape)

    help_atom_shape = 'The **size of each atom** dimension. To use the ground truth atom size, tick the checkbox above.'
    atom_shape = tuple([st.sidebar.number_input('Atom size',
                        value=default_params['atom_shape'][0], min_value=1,
                        help=help_atom_shape)] * len(default_params['atom_shape'])) if not use_atom_shape else None
    if verbose and not use_atom_shape:
        st.sidebar.caption(help_atom_shape)

    help_sparsity_H = 'The strength of the **L1 activation sparsity regularization** imposed on the optimization problem.'
    sparsity_H = st.sidebar.number_input('Activation sparsity', min_value=0.0, value=0.0, step=0.01,
                                         help=help_sparsity_H)
    if verbose:
        st.sidebar.caption(help_sparsity_H)

    help_inhibition_strength = \
        '''The strength of the **lateral activation sparsity regularization** imposed on the optimization problem. The
        parameter controls how strong the activation of an atom at a particular shift location suppresses the activation of
        the same atom at neighboring locations.'''
    inhibition_strength = st.sidebar.number_input('Lateral activation inhibition', min_value=0.0, value=0.1, step=0.01,
                                                  help=help_inhibition_strength)
    if verbose:
        st.sidebar.caption(help_inhibition_strength)

    help_n_iterations = '''The **number of multiplicative updates** to the atom dictionary and activation tensors.'''
    n_iterations = st.sidebar.number_input('# Iterations', value=100, min_value=1, help=help_n_iterations)
    if verbose:
        st.sidebar.caption(help_n_iterations)

    help_backend = \
        '''The **optimization backend** for computing the multiplicative gradients.
            **Note:** All backends yield the same results. Within the scope of this demo, switching between backends is thus
            for speed comparisons only.'''
    backend = st.sidebar.selectbox('Backend', ['numpy', 'numpy_fft', 'numpy_caching_fft', 'pytorch', 'pytorch_fft'], 4,
                                   help=help_backend)
    if verbose:
        st.sidebar.caption(help_backend)

    help_reconstruction_mode = \
        '''Defines the **convolution mode** for the signal reconstruction.\

        **valid:** The activation tensor
        is smaller than the input by the atom size along the shift dimensions, so that the convolution of atoms and
        activations matches the size of the input.\

        **full:** The activation tensor is larger than the input by the atom size along the shift dimensions. Compared to the
        'valid' reconstruction mode, this also creates shifted versions of each atom that only partially overlap with the input
        array. The convolution result is trimmed to the appropriate size.\

        **circular:** The activation tensor is of the same
        size as the input. Other than in 'full' mode, parts of the convolution result that are outside the range of the
        input array are inserted circularly on the respective other side of the array.'''
    reconstruction_mode = st.sidebar.selectbox('Reconstruction', ['valid', 'full', 'circular'], 2,
                                               help=help_reconstruction_mode)
    if verbose:
        st.sidebar.caption(help_reconstruction_mode)

    selected_params = dict(
        n_iterations=n_iterations,
        n_atoms=n_atoms,
        atom_shape=atom_shape,
        sparsity_H=sparsity_H,
        inhibition_strength=inhibition_strength,
        backend=backend,
        reconstruction_mode=reconstruction_mode,
    )

    # default parameter selection dictionary
    use_default = dict(
        n_atoms=use_n_atoms,
        atom_shape=use_atom_shape,
    )

    # overwrite selected parameters with defaults
    selected_params.update({k: v for k, v in default_params.items() if use_default[k]})

    return selected_params


class SignalTool(ABC):
    """An abstract base class that serves as a factory for creating specialized objects that facilitate the handling of
    different signal types."""

    def __new__(cls, n_dims: int):
        """
        Parameters
        ----------
        n_dims : int
            The dimensionality of the signals to be managed.
            * n_dims=1 for time series.
            * n_dims=2 for image data.
        """
        if n_dims == 1:
            return super(SignalTool, cls).__new__(SignalTool1D)
        if n_dims == 2:
            return super(SignalTool, cls).__new__(SignalTool2D)
        raise ValueError("'n_dims' must be in {1, 2}")

    @classmethod
    def st_generate_input(cls, verbose: bool = True) -> Tuple[np.ndarray, dict]:
        """
        Defines all signal parameters via streamlit widgets and returns a generated input matrix V for the NMF together
        with a dictionary containing details of the used NMF atoms.

        Parameters
        ----------
        verbose : bool
            If True, show detailed information.

        Returns
        -------
        V : np.ndarray
            The generated input for the NMF.
        nmf_params : dict
            Ground truth NMF atom parameters that were used for the signal generation.
        """
        st.sidebar.markdown('# Signal settings')

        # define the number input signals
        help_n_signals = \
            '''The **number of generated signals** passed as input to the algorithm.
            All signals have the same shape and number of channels.'''
        n_signals = st.sidebar.number_input('# Signals', min_value=1, value=10, help=help_n_signals)
        if verbose:
            st.sidebar.caption(help_n_signals)
        signal_params = cls.st_define_signal_params(verbose=verbose)

        # create the input
        V = []
        for _ in range(n_signals):
            signal, W = cls.generate_signal(signal_params)
            V.append(signal)
        V = np.stack(V)

        # extract the ground truth NMF parameters
        nmf_params = {'n_atoms': W.shape[0], 'atom_shape': W.shape[2:]}

        return V, nmf_params

    @classmethod
    def st_compare_signals(cls, V: np.ndarray, R: np.ndarray, verbose: bool = True):
        """
        Compares a given input matrix with its NMF reconstruction in streamlit.

        Parameters
        ----------
        V : np.ndarray
            The input that was factorized via NMF.
        R : np.ndarray
            The NMF reconstruction of the input.
        verbose : bool
            If True, show detailed information.
        """
        st.markdown('# Global signal reconstruction')

        # show explanatory text
        if verbose:
            st.caption('''The visualization below gives a **first impression of the reconstruction accuracy** of the
            learned factorization.\

            **Left**: The entire input provided to the algorithm visualized as a matrix. Each row represents a particular input
                signal, whose channels and remaining dimensions have been stacked into a single vector.\

            **Middle**: The reconstruction of the signal obtained through the learned factorization, reshaped in the same way.\

            **Right**: The difference between the two matrices.
            ''')

        # show the input, its reconstruction, and the reconstruction error as images next to each other
        cols = st.beta_columns(3)
        for col, X, title in zip(cols, [V, R, V-R], ['Input', 'Reconstruction', 'Error']):
            with col:
                fig = plt.figure()
                plt.imshow(X.reshape(X.shape[0], -1), aspect='auto', interpolation='none')
                plt.title(title)
                st.pyplot(fig)

    @classmethod
    def st_compare_individual_signals(cls, V: np.ndarray, R: np.ndarray, verbose: bool = True):
        """
        Selects a particular signal and its reconstruction from the given input via a streamlit widget and compares them.

        Parameters
        ----------
        V : np.ndarray
            The input that was factorized via NMF.
        R : np.ndarray
            The NMF reconstruction of the input.
        verbose : bool
            If True, show detailed information.
        """
        st.markdown('# Individual signal reconstruction')

        # show explanatory text
        if verbose:
            st.caption('''The visualization below shows a **comparison between an individual input signal and its
            reconstruction** obtained through the learned factorization model. If more than one input signal has been
            generated, the signal to be visualized can be selected via the **slider**.''')

        # select and compare signals
        i_signal = st.slider('Signal number', 1, V.shape[0]) - 1 if V.shape[0] > 1 else 0
        cls._st_compare_individual_signals(V[i_signal], R[i_signal])

    @classmethod
    def st_plot_partial_reconstructions(cls, V: np.ndarray, nmf: TransformInvariantNMF, verbose: bool = True):
        """
        Visualizes the partial reconstructions of the given input by the different NMF atoms.

        Parameters
        ----------
        V : np.ndarray
            The input that was factorized via NMF.
        nmf : TransformInvariantNMF
            The trained NMF object.
        verbose : bool
            If True, show detailed information.
        """
        st.markdown('# Partial signal reconstructions')

        # show explanatory text
        if verbose:
            st.caption('''The visualization below shows the **learned dictionary atoms (left) and their partial contributions
            (right)** to the reconstruction of an individual signal. If more than one input signal has been generated, the
            signal to be visualized can be selected via the **slider**.''')

        # select the signal and show the partial reconstruction
        i_signal = st.slider('Signal number', 1, V.shape[0], key='i_signal_partial') - 1 if V.shape[0] > 1 else 0
        for i_atom in range(nmf.n_atoms):
            R_atom = nmf.R_partial(i_atom)
            cols = st.beta_columns(2)
            for col, signals, signal_opts, opts in zip(
                    cols,
                    [[nmf.W[i_atom]], [R_atom[i_signal], V[i_signal]]],
                    [[{}], [{'label': 'Atom contribution', 'color': 'tab:red', 'linestyle': '--'}, {'label': 'Signal'}]],
                    [{'title': f'Atom {i_atom + 1}'}, {'title': 'Atom contribution'}],
            ):
                with col:
                    cls.plot_signals(signals, signal_opts, opts)

    @classmethod
    def plot_signals(
            cls,
            signals: List[np.ndarray],
            signal_opts: Optional[Iterable[dict]] = None,
            opts: Optional[dict] = None,
    ):
        """Wrapper for `_plot_signals` to fill the default arguments."""
        if signal_opts is None:
            signal_opts = repeat({})
        if opts is None:
            opts = {}
        cls._plot_signals(signals, signal_opts, opts)

    @classmethod
    @abstractmethod
    def st_define_signal_params(cls, verbose: bool = True) -> dict:
        """
        Defines all signal parameters via streamlit widgets and returns them in a dictionary.

        Parameters
        ----------
        verbose : bool
            If True, show detailed information.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def generate_signal(cls, signal_params: dict) -> Tuple[np.ndarray, np.ndarray]:
        """Creates a single signal using the specified signal parameters. Returns the signal and the used NMF atoms."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def _st_compare_individual_signals(cls, V_i: np.ndarray, R_i: np.ndarray):
        """
        Compares a single signal and its reconstruction in streamlit.

        Parameters
        ----------
        V_i : np.ndarray
            A single input signal from the input that was factorized via NMF.
        R_i : np.ndarray
            The NMF reconstruction of the given input signal.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def _plot_signals(cls, signals: List[np.ndarray], signal_opts: Iterable[dict], opts: dict):
        """
        Visualizes a given list of signals.

        Parameters
        ----------
        signals : List[np.ndarray]
            The list of signals to be plotted.
        signal_opts : Iterable[dict]
            A list of dictionaries containing plotting options for each individual signal.
        opts : dict
            A dictionary containing global plotting options.
        """
        raise NotImplementedError


class SignalTool1D(SignalTool):
    """A utility class to handle 1-D multi-channel signals (time series data)."""

    @classmethod
    def st_define_signal_params(cls, verbose: bool = True) -> dict:

        # define the number of channels
        n_channels = st.sidebar.number_input('# Channels', min_value=1, value=3, max_value=5, help=HELP_CHANNEL)
        if verbose:
            st.sidebar.caption(HELP_CHANNEL)

        # define the pulse shapes
        help_shapes = \
            '''The **pulse shapes** that can be combined along different channels to form a multi-channel time-series
            symbol.'''
        shapes = st.sidebar.multiselect('Pulse shapes', ['n', '-', '^', 'v', '_'], ['n', '-', '^', 'v', '_'], help=help_shapes)
        if verbose:
            st.sidebar.caption(help_shapes)

        # to avoid having too many different symbols, consider only those where all channels are identical
        help_symbols = \
            '''The **set of symbols** that form the atom dictionary of of multi-channel time series symbols for
             signal generation. The subset of channel-symmetric symbols is pre-selected.'''
        symbols = st.sidebar.multiselect('Symbols', [''.join(chars) for chars in product(*repeat(shapes, n_channels))],
                                         [s * n_channels for s in shapes], help=help_symbols)
        if verbose:
            st.sidebar.caption(help_symbols)

        # define the number of pulses
        help_n_pulses = 'The **number of symbols** concatenated randomly to form a signal.'
        n_pulses = st.sidebar.number_input('# Symbols', min_value=1, value=3, help=help_n_pulses)
        if verbose:
            st.sidebar.caption(help_n_pulses)

        # define the pulse length
        help_pulse_length = 'The **length** of each individual multi-channel symbol.'
        pulse_length = st.sidebar.number_input('Symbol length', min_value=1, value=20, help=help_pulse_length)
        if verbose:
            st.sidebar.caption(help_pulse_length)

        # create the parameter dictionary
        signal_params = dict(
            n_pulses=n_pulses,
            symbols=symbols,
            pulse_length=pulse_length,
        )

        return signal_params

    @classmethod
    def generate_signal(cls, signal_params: dict) -> Tuple[np.ndarray, np.ndarray]:
        signal, W = generate_pulse_train(**signal_params)
        return signal, W

    @classmethod
    def _st_compare_individual_signals(cls, V_i: np.ndarray, R_i: np.ndarray):
        signals = [V_i, R_i]
        opts = [{'label': 'Reconstruction', 'color': 'tab:red', 'linestyle': '--'}, {'label': 'Signal', 'zorder': -1}]
        cls.plot_signals(signals, opts)

    @classmethod
    def _plot_signals(cls, signals: List[np.ndarray], signal_opts: Iterable[dict], opts: dict):
        assert len(np.unique([signal.shape[0] for signal in signals])) == 1
        n_channels = signals[0].shape[0]
        with plt.style.context('seaborn'):
            fig, axs = plt.subplots(nrows=n_channels)
            axs = np.atleast_1d(axs)
            for signal, signal_opt in zip(signals, signal_opts):
                for i_channel, ax in enumerate(axs):
                    ax.plot(signal[i_channel], **signal_opt)
            plt.legend()
            fig.suptitle(opts.get('title'))
            st.pyplot(fig)


class SignalTool2D(SignalTool):
    """A utility class to handle 2-D multi-channel signals (image data)."""

    @classmethod
    def st_define_signal_params(cls, verbose: bool = True) -> dict:

        # choose between grayscale or color images
        n_channels = st.sidebar.radio('# Channels', ['1 (Grayscale images)', '3 (Color images)'], 0, help=HELP_CHANNEL)
        n_channels = 1 if n_channels == '1 (Grayscale images)' else 3
        if verbose:
            st.sidebar.caption(HELP_CHANNEL)

        # create all possible combinations of shapes and color
        shapes = ['+', 'x', 's']
        colors = ['r', 'g', 'b', 'y', 'm', 'c', 'w'] if n_channels == 3 else ['']
        all_symbols = [''.join(spec) for spec in product(shapes, colors)]

        # select the symbols to be used
        help_symbols = \
            '''The **set of symbols** that form the atom dictionary of image patches for signal generation. The first
            character defines the visual shape of the corresponding symbol. The second character (only available for color
            images) defines the color of the symbol. For color images, a representative subset of symbols is pre-selected.'''
        symbols = st.sidebar.multiselect('Symbols', all_symbols, ['+r', 'xg', 'sb'] if n_channels == 3 else all_symbols,
                                         help=help_symbols)
        if verbose:
            st.sidebar.caption(help_symbols)

        # select the number of symbols per dimension
        help_n_symbols = \
            '''The **number of symbols** placed randomly next to each other both horizontally and vertically to generate an
            input image.'''
        n_symbols = st.sidebar.number_input('# Symbols per dimension', min_value=1, value=5, help=help_n_symbols)
        if verbose:
            st.sidebar.caption(help_n_symbols)

        # define the size of each symbol
        help_symbol_size = 'The **size of each symbol** in pixels per dimension. All symbols are of square size.'
        symbol_size = st.sidebar.number_input('Symbol size', min_value=1, value=10, help=help_symbol_size)
        if verbose:
            st.sidebar.caption(help_symbol_size)

        # create the parameter dictionary
        signal_params = dict(
            n_symbols=n_symbols,
            symbol_size=symbol_size,
            symbols=symbols,
        )

        return signal_params

    @classmethod
    def generate_signal(cls, signal_params: dict) -> Tuple[np.ndarray, np.ndarray]:
        signal, W = generate_block_image(**signal_params)
        return signal, W

    @classmethod
    def _st_compare_individual_signals(cls, V_i: np.ndarray, R_i: np.ndarray):
        cols = st.beta_columns(2)
        for col, X, title in zip(cols, [V_i, R_i], ['Input', 'Reconstruction']):
            with col:
                cls.plot_signals([X], opts={'title': title})

    @classmethod
    def _plot_signals(cls, signals: List[np.ndarray], signal_opts: Iterable[dict], opts: dict):
        fig = plt.figure()
        plt.imshow(signals[0].transpose((1, 2, 0)) / signals[0].max())
        plt.title(opts.get('title'))
        st.pyplot(fig)



"""
=================================================================
Convergence control and iteration abort methods
=================================================================
This examples demonstrates how to log and plot the reconstruction
energy and use its value to abort iteration once a given
convergence criterion is fulfilled.
"""
import numpy as np
import matplotlib.pyplot as plt

from tnmf.TransformInvariantNMF import TransformInvariantNMF
from tnmf.utils.data_loading import racoon_image

print(__doc__)

# Load an example image.
img = racoon_image(scale=0.1)

# Construct a TransformInvariantNMF instance with selected parameters for the model.
n_atoms = 16
nmf = TransformInvariantNMF(
    n_atoms=n_atoms,
    atom_shape=(6, 10),
    n_iterations=500,
    reconstruction_mode='circular',
    backend='pytorch',
    verbose=3,
)

# Samples provided to nmf.fit() have to be indexed as V[sample, channel, sample_dim_1 .. sample_dim_n].
# This example uses one single-channel image.
img = img[np.newaxis, np.newaxis, ...]

# Define a progress callback to keep track of the reconstruction energy in every iteration.
reconstruction_energy = list()


def progress_callback(nmf_instance: TransformInvariantNMF, iteration: int) -> bool:
    energy = nmf_instance._energy_function(img)  # TODO: having this function protected is impractical
    reconstruction_energy.append([iteration, energy])

    # Continue iteration as long as energy is above a certain threshold.
    return energy > 20.


# Run the fitting, i.e. compute dictionary W and activations H so that img = H*W.
# Note that setting a progress callback suppresses regular convergence output.
nmf.fit(img, inhibition_strength=0.0, progress_callback=progress_callback)

# Collect results from the TransformInvariantNMF instance.
img_r = nmf.R

# Create a plot of the original image and the reconstruction.
fig, axes = plt.subplots(nrows=1, ncols=2, squeeze=False, figsize=(8, 4))
for (data, title, kwargs), ax in zip([
        (img[0, 0], 'Original', dict(cmap='Greys')),
        (img_r[0, 0], 'Reconstruction', dict(cmap='Greys')),
        ], axes.flatten()):
    ax.set_title(title)
    ax.imshow(data, **kwargs, vmin=0., vmax=1.)
plt.tight_layout()
plt.show()

# Create a plot of the reconstruction energy over iterations.
reconstruction_energy = np.array(reconstruction_energy).T
plt.figure(figsize=(6, 4))
plt.plot(*reconstruction_energy, '-o')
plt.grid()
plt.xlabel('Iteration')
plt.ylabel('Reconstruction Energy')
plt.show()



# pylint: skip-file
try:
    from importlib.metadata import version, PackageNotFoundError
except ImportError:  # for Python<3.8
    from importlib_metadata import version, PackageNotFoundError

try:
    __version__ = version("tnmf")
except PackageNotFoundError:
    pass

del version
del PackageNotFoundError



"""
Test if `from tnmf import *` does not fail
"""

# pylint: disable=broad-except, wildcard-import, unused-wildcard-import

_import_error = None

try:
    from tnmf import *
except Exception as e:
    _import_error = e


def test_import_tinmf():
    assert _import_error is None



# pylint: disable=missing-module-docstring
import numpy as np

from scipy.misc import face
from PIL import Image


def racoon_image(gray: bool = True, scale: float = 1.):
    img = Image.fromarray(face(gray))
    img = img.resize([int(scale * s) for s in img.size])
    img = np.array(img, dtype=float) / 255
    return img



import importlib.resources
from importlib import import_module

import numpy as np
import streamlit as st

# show TNMF header
with importlib.resources.path('logos', 'tnmf_header.png') as img_file:
    st.image(str(img_file), use_column_width='always')

# mapping of streamlit demo names to the corresponding demo file and optional parameters
DEMO_NAME_DICT = {
    '1-D Synthetic Signals': ('synthetic_signals', {'n_dims': 1}),
    '2-D Synthetic Signals': ('synthetic_signals', {'n_dims': 2}),
}

# create progress bar on the top that is shown for all demos
progress_bar = st.sidebar.progress(1.)

# toggle verbose mode
help_verbose = 'Displays / hides **detailed explanations**.'
verbose = st.sidebar.checkbox('Verbose', True, help=help_verbose)
if verbose:
    st.sidebar.caption(help_verbose)

# select the demo
help_select_demo = 'The specific **demo example** that gets executed.'
selected_demo = st.sidebar.selectbox('Demo example', list(DEMO_NAME_DICT.keys()), 1, help=help_select_demo)
if verbose:
    st.sidebar.caption(help_select_demo)

# select the random seed
help_seed = 'The fixed **random seed** that is used for the simulation.'
seed = st.sidebar.number_input('Random seed', value=42, help=help_seed)
np.random.seed(seed)
if verbose:
    st.sidebar.caption(help_seed)

# extract the demo name and parameters, import the corresponding module, and execute the demo
demo_name, demo_args = DEMO_NAME_DICT[selected_demo]
demo_module = import_module(demo_name)
demo_module.main(progress_bar, verbose=verbose, **demo_args)



"""
A module that provides some specializations and utilities for all PyTorch based backends.
"""

# TODO: it should be possible to reformulate the gradients using
#       https://pytorch.org/docs/stable/autograd.html#torch.autograd.functional.jacobian
# TODO: merge gradient functions into one
# TODO: add device option

from typing import Tuple, Optional, Union
from itertools import chain

import numpy as np
import torch
from torch import Tensor
from torch.nn.functional import conv1d

from ._Backend import Backend


# pylint: disable=abstract-method
class PyTorchBackend(Backend):
    r"""
    The parent class for all PyTorch based backends.

    They provide the functionality to evaluate the gradients of the factorization model via automatic differentiation
    using :mod:`torch.autograd`.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._padding = None

    def _initialize_matrices(
        self,
        V: np.ndarray,
        atom_shape: Tuple[int, ...],
        n_atoms: int,
        W: Optional[Tensor] = None,
    ) -> Tuple[Tensor, Tensor]:

        if self._reconstruction_mode == 'valid':
            self._padding = None
        elif self._reconstruction_mode == 'full':
            pad_shape = tuple(chain.from_iterable((a - 1, a - 1) for a in reversed(atom_shape)))
            self._padding = dict(pad=pad_shape, mode='constant', value=0)
        elif self._reconstruction_mode in ('circular', 'reflect'):
            pad_shape = tuple(chain.from_iterable((a - 1, 0) for a in reversed(atom_shape)))
            self._padding = dict(pad=pad_shape, mode=self._reconstruction_mode)
        else:
            raise ValueError(f'Unsupported reconstruction mode "{self._reconstruction_mode}".'
                             f'Please choose "valid", "full", "circular", or "reflect".')

        w, h = super()._initialize_matrices(V, atom_shape, n_atoms, W)

        H = torch.from_numpy(h)
        if W is None:
            W = torch.from_numpy(w)

        return W, H

    @staticmethod
    def to_ndarray(arr: Tensor) -> np.ndarray:
        return arr.detach().numpy()

    @staticmethod
    def convolve_multi_1d(arr: Tensor, kernels: Tuple[np.ndarray, ...], axes: Tuple[int, ...]) -> Tensor:
        assert len(kernels) == len(axes)

        convolved = arr.detach().clone()
        for a, kernel in zip(axes, kernels):
            pad = (len(kernel) - 1) // 2
            kernel_view = torch.as_tensor(kernel).view(1, 1, -1)
            # axis to be convolve across has to be the last one
            convolved = torch.movedim(convolved, a, -1)
            convolved_shape = convolved.size()
            # all other non-singleton axes have to be aggregated into the batch_size axis (0th axis)
            convolved = torch.reshape(convolved, (-1, 1, convolved.size()[-1]))
            convolved = conv1d(convolved, kernel_view, padding=pad)
            # restore original shape
            convolved = torch.reshape(convolved, convolved_shape)
            # restore original axis order
            convolved = torch.movedim(convolved, -1, a)

        return convolved

    @staticmethod
    def normalize(arr: Tensor, axis: Optional[Union[int, Tuple[int, ...]]] = None):
        arr.divide_(arr.sum(dim=axis, keepdim=True))

    def reconstruction_gradient_W(self, V: np.ndarray, W: Tensor, H: Tensor) -> Tuple[Tensor, Tensor]:
        W_grad = W.detach().requires_grad_()
        neg_energy, pos_energy = self._energy_terms(V, W_grad, H)
        neg = torch.autograd.grad(neg_energy, W_grad, retain_graph=True)[0]
        pos = torch.autograd.grad(pos_energy, W_grad)[0]
        return neg.detach(), pos.detach()

    def reconstruction_gradient_H(self, V: np.ndarray, W: Tensor, H: Tensor) -> Tuple[Tensor, Tensor]:
        H_grad = H.detach().requires_grad_()
        neg_energy, pos_energy = self._energy_terms(V, W, H_grad)
        neg = torch.autograd.grad(neg_energy, H_grad, retain_graph=True)[0]
        pos = torch.autograd.grad(pos_energy, H_grad)[0]
        return neg.detach(), pos.detach()

    def _energy_terms(self, V: np.ndarray, W: Tensor, H: Tensor) -> Tuple[Tensor, Tensor]:
        V = torch.as_tensor(V)
        R = self.reconstruct(W, H)
        neg = (R * V).sum()
        pos = 0.5 * (V.square().sum() + R.square().sum())
        return neg, pos

    def reconstruction_energy(self, V: Tensor, W: Tensor, H: Tensor) -> float:
        V = torch.as_tensor(V)
        R = self.reconstruct(W, H)
        energy = 0.5 * torch.sum(torch.square(V - R))
        return float(energy)



"""
Transform-Invariant Non-Negative Matrix Factorization

Authors: Adrian Šošić, Mathias Winkel
"""

# TODO: backend-specific return types
# TODO: no atom_shape for non-ShiftInvariantNMF
# TODO: handling of transform input/output shapes
# TODO: naming convention: energy (instead of cost/error)
# TODO: add options for tensor renormalization
# TODO: cache reconstruction result
# TODO: flexible input types for V
# TODO: we extract .shape[...] too often
# TODO: add support for inhibition

import logging
from typing import Tuple, Callable, Union

import numpy as np

from .backends.NumPy import NumPy_Backend
from .backends.NumPy_FFT import NumPy_FFT_Backend
from .backends.PyTorch import PyTorch_Backend
from .backends.PyTorch_FFT import PyTorch_FFT_Backend
from .backends.NumPy_CachingFFT import NumPy_CachingFFT_Backend


class TransformInvariantNMF:
    r"""
    Transform Invariant Non-Negative Matrix Factorization.

    Finds non-negative tensors :attr:`W` (dictionary) and :attr:`H` (activations) that approximate the non-negative tensor
    :attr:`V` (samples) for a given transform operator`. # TODO: add link to TNMF model

    .. note::
        Currently, only a single transform type, corresponding to *shift invariance*, is supported and hard-coded. In contrast
        to other *generic* types of transforms, shift invariance can be efficiently achieved through convolution operations
        (or, equivalently, multiplication in Fourier domain). Therefore, shift invariance will remain hard-coded and retained
        as an optional transform type even when additional transforms become supported in future releases.

    Optimization is performed via multiplicative updates to :attr:`W` and :attr:`H`, see [1]_.
    Different backend implementations (NumPy, PyTorch, with/without FFT, etc.) can be selected by the user.

    Parameters
    ----------
    n_atoms : int
        Number of elementary atoms. The shape of :attr:`W` will be ``(n_atoms, n_channels, *atom_shape)``.
    atom_shape : Tuple[int, ...]
        Shape of the elementary atoms. The shape of :attr:`W` will be ``(n_atoms, n_channels, *atom_shape)``.
    inhibition_range: Union[int, Tuple[int, ...]], default = None
        Lateral inhibition range. If set to None, the value is set to ``*atom_shape``, which ensures
        that activations are pairwise sufficiently far apart, that the corresponding atoms do not overlap
        in the reconstruction.
    n_iterations : int, default = 1000
        Maximum number of iterations (:attr:`W` and :attr:`H` updates) to be performed.
    backend : {'numpy', 'numpy_fft', 'numpy_caching_fft', 'pytorch'}, default = 'numpy_fft'
        Defines the optimization backend.

        * `'numpy'`: Selects the :class:`.NumPy_Backend`.
        * `'numpy_fft'` (default): Selects the :class:`.NumPy_FFT_Backend`.
        * `'numpy_caching_fft'`: Selects the :class:`.NumPy_CachingFFT_Backend`.
        * `'pytorch'`: Selects the :class:`.PyTorch_Backend`.
        * `'pytorch_fft'`: Selects the :class:`.PyTorch_FFT_Backend`.

    logger : logging.Logger, default = None
        Logger instance used for intermediate output. If None, an internal logger instance will be created.
    verbose : {0, 1, 2, 3}, default = 0
        Verbosity level.

        * 0: Show only errors.
        * 1: Include warnings.
        * 2: Include info.
        * 3: Include debug messages.

    **kwargs
        Keyword arguments that are handed to the constructor of the backend class.

    Attributes
    ----------
    W : np.ndarray
        The dictionary tensor of shape ``(n_atoms, num_channels, *atom_shape)``.
    H : np.ndarray
        The activation tensor of shape ``(num_samples, num_atoms, *shift_shape)``.
    R : np.ndarray
        The reconstruction of the sample tensor using the current :attr:`W` and :attr:`H`.

    Examples
    --------
        TODO: add examples

    Notes
    -----
    Planned features:

        * Batch processing
        * Arbitrary transform types
        * Nested transformations
        * Additional reconstruction norms

    References
    ----------
    # TODO: add bibtex file

    .. [1] Lee, D.D., Seung, H.S., 2000. Algorithms for Non-negative Matrix Factorization,
        in: Proceedings of the 13th International Conference on Neural Information
        Processing Systems. pp. 535–541. https://doi.org/10.5555/3008751.3008829
    """

    def __init__(
            self,
            n_atoms: int,
            atom_shape: Tuple[int, ...],
            inhibition_range: Union[int, Tuple[int, ...]] = None,
            n_iterations: int = 1000,
            backend: str = 'numpy_fft',
            logger: logging.Logger = None,
            verbose: int = 0,
            **kwargs,
    ):
        self.atom_shape = atom_shape

        if inhibition_range is None:
            # default inhibition range = minimal range to cover the atom size
            self._inhibition_range = tuple(a-1 for a in atom_shape)
        elif isinstance(inhibition_range, int):
            self._inhibition_range = (inhibition_range, ) * len(atom_shape)
        else:
            self._inhibition_range = inhibition_range

        assert len(self._inhibition_range) == len(atom_shape)
        self._inhibition_kernels_1D = tuple((1 - ((np.arange(-i, i + 1) / (i+1)) ** 2) for i in self._inhibition_range))
        self.n_atoms = n_atoms
        self.n_iterations = n_iterations
        self._axes_W_normalization = tuple(range(-len(atom_shape), 0))
        self.eps = 1.e-9

        backend_map = {
            'numpy': NumPy_Backend,
            'numpy_fft': NumPy_FFT_Backend,
            'numpy_caching_fft': NumPy_CachingFFT_Backend,
            'pytorch': PyTorch_Backend,
            'pytorch_fft': PyTorch_FFT_Backend,
        }

        self._backend = backend_map[backend.lower()](**kwargs)

        self._logger = logger if logger is not None else logging.getLogger(self.__class__.__name__)
        self._logger.setLevel([logging.ERROR, logging.WARNING, logging.INFO, logging.DEBUG][verbose])
        self._logger.debug(f'Using {backend} backend.')

        self.n_iterations_done = 0
        self._W = None
        self._H = None

    @property
    def W(self) -> np.ndarray:
        return self._backend.to_ndarray(self._W)

    @property
    def H(self) -> np.ndarray:
        return self._backend.to_ndarray(self._H)

    @property
    def R(self) -> np.ndarray:
        return self._backend.to_ndarray(self._reconstruct())

    def R_partial(self, i_atom: int) -> np.ndarray:
        return self._backend.to_ndarray(self._backend.partial_reconstruct(self._W, self._H, i_atom))

    def _reconstruct(self) -> np.ndarray:
        return self._backend.reconstruct(self._W, self._H)

    def _energy_function(self, V: np.ndarray) -> float:
        return self._backend.reconstruction_energy(V, self._W, self._H)

    def _multiplicative_update(self, arr: np.ndarray, neg, pos, sparsity: float = 0.):
        assert sparsity >= 0

        regularization = self.eps

        if sparsity > 0:
            regularization += sparsity

        pos += regularization

        arr *= neg
        arr /= pos

    def _update_W(self, V: np.ndarray):
        neg, pos = self._backend.reconstruction_gradient_W(V, self._W, self._H)
        self._multiplicative_update(self._W, neg, pos)
        self._backend.normalize(self._W, axis=self._axes_W_normalization)

    def _update_H(self, V: np.ndarray, sparsity: float = 0., inhibition: float = 0.):
        # TODO: sparsity and inhibition should be handled by the backends
        neg, pos = self._backend.reconstruction_gradient_H(V, self._W, self._H)

        # add the inhibition gradient component
        if inhibition > 0:
            # TODO: maybe also add cross-channel/cross-atom inhibition?
            convolve_axes = range(-len(self.atom_shape), 0)
            inhibition_gradient = self._backend.convolve_multi_1d(self._H, self._inhibition_kernels_1D, convolve_axes)
            inhibition_gradient -= self._H
            inhibition_gradient *= inhibition
            pos += inhibition_gradient

        self._multiplicative_update(self._H, neg, pos, sparsity)

    def _do_fit(
            self,
            V: np.ndarray,
            update_H: bool,
            update_W: bool,
            sparsity_H: float,
            inhibition_strength: float,
            keep_W: bool,
            progress_callback: Callable[['TransformInvariantNMF', int], bool],
    ):
        assert update_H or update_W
        assert sparsity_H >= 0
        assert inhibition_strength >= 0

        self._W, self._H = self._backend.initialize(
            V, self.atom_shape, self.n_atoms, self._W if keep_W else None)

        if not keep_W:
            self._backend.normalize(self._W, self._axes_W_normalization)

        for self.n_iterations_done in range(self.n_iterations):
            if update_H:
                self._update_H(V, sparsity_H, inhibition_strength)

            if update_W:
                self._update_W(V)

            if progress_callback is not None:
                if not progress_callback(self, self.n_iterations_done):
                    break
            else:
                self._logger.info(f"Iteration: {self.n_iterations_done}\tEnergy function: {self._energy_function(V)}")

        self._logger.info("NMF finished.")

    def fit(
            self,
            V: np.ndarray,
            update_H: bool = True,
            update_W: bool = True,
            sparsity_H: float = 0.,
            inhibition_strength: float = 0.,
            progress_callback: Callable[['TransformInvariantNMF', int], bool] = None,
    ):
        r"""
        Perform non-negative matrix factorization of samples :attr:`V`, i.e. optimization of dictionary :attr:`W` and
        activations :attr:`H`.

        Parameters
        ----------
        V : np.ndarray
            Samples to be reconstructed. The shape of the sample tensor is ``(n_samples, n_channels, *sample_shape)``,
            where `sample_shape` is the shape of the individual samples and each sample consists of `n_channels`
            individual channels.
        update_H : bool, default = True
            If False, the activation tensor :attr:`H` will not be updated.
        update_W : bool, default = True
            If False, the dictionary tensor :attr:`W' will not be updated.
        sparsity_H : float, default = 0.
            Sparsity enforcing regularization for the :attr:`H` update.
        inhibition_strength : float, default = 0.
            Lateral inhibition regularization factor for the :attr:`H` update.
        progress_callback : Callable[['TransformInvariantNMF', int], bool], default = None
            If provided, this function will be called after every iteration, i.e. after every update to :attr:`H` and
            :attr:`W`. The first parameter to the function is the calling :class:`TransformInvariantNMF` instance, which can be
            used to inspect intermediate results, etc. The second parameter is the current iteration step.

            If the `progress_callback` function returns False, iteration will be aborted, which allows to implement
            specialized convergence criteria.
        """
        self._do_fit(V, update_H, update_W, sparsity_H, inhibition_strength, False, progress_callback)

    def partial_fit(
            self,
            V: np.ndarray,
            update_H: bool = True,
            update_W: bool = True,
            sparsity_H: float = 0.,
            inhibition_strength: float = 0.,
            progress_callback: Callable[['TransformInvariantNMF', int], bool] = None,
    ):
        self._do_fit(V, update_H, update_W, sparsity_H, inhibition_strength, self.n_iterations_done > 0, progress_callback)



"""
Test the decomposition on two identical images with different values for sparsity and inhibition and
ensure that all backends yield the same reconstruction energy and L0 and L1 norm for the activations H
"""

import logging
from typing import Dict

import numpy as np
import pytest

from tnmf.TransformInvariantNMF import TransformInvariantNMF
from tnmf.utils.data_loading import racoon_image

logging.basicConfig(level=logging.DEBUG, format='%(asctime)s - %(levelname)s: %(message)s')

# define all test settings
backends = ['numpy', 'numpy_fft', 'numpy_caching_fft', 'pytorch', 'pytorch_fft']
# hard-coded expected energy levels for the different reconstruction modes
test_params_sparsity = [
    # increasing sparsity, no inhibition
    {'fit': dict(sparsity_H=0.0), 'nmf': dict(), 'res': dict(energy=186.666013, norm_H_1=7704.38977, norm_H_0=176346)},
    {'fit': dict(sparsity_H=0.1), 'nmf': dict(), 'res': dict(energy=225.494731, norm_H_1=6563.91176, norm_H_0=174037)},
    {'fit': dict(sparsity_H=0.5), 'nmf': dict(), 'res': dict(energy=858.621063, norm_H_1=4258.82541, norm_H_0=155247)},
    {'fit': dict(sparsity_H=1.0), 'nmf': dict(), 'res': dict(energy=2429.69334, norm_H_1=2114.50047, norm_H_0=136396)},
    {'fit': dict(sparsity_H=5.0), 'nmf': dict(), 'res': dict(energy=5351.91865, norm_H_1=3.0800e-06, norm_H_0=65338)},
    {'fit': dict(sparsity_H=10.), 'nmf': dict(), 'res': dict(energy=5351.91866, norm_H_1=2.5103e-13, norm_H_0=62486)},
]
test_params_inhibition = [
    # no sparsity, increasing inhibition
    {'fit': dict(inhibition_strength=0.1), 'nmf': dict(), 'res': dict(energy=435.993252, norm_H_1=5574.6769, norm_H_0=175483)},
    {'fit': dict(inhibition_strength=0.5), 'nmf': dict(), 'res': dict(energy=1831.92669, norm_H_1=3031.4130, norm_H_0=168931)},
    {'fit': dict(inhibition_strength=1.0), 'nmf': dict(), 'res': dict(energy=2665.70594, norm_H_1=2252.1331, norm_H_0=160644)},
    {'fit': dict(inhibition_strength=5.0), 'nmf': dict(), 'res': dict(energy=3618.33671, norm_H_1=1947.1969, norm_H_0=129929)},
    {'fit': dict(inhibition_strength=10.), 'nmf': dict(), 'res': dict(energy=3779.64954, norm_H_1=1926.9570, norm_H_0=118795)},
    # no sparsity, increasing inhibition, smaller inhibition range
    # pylint: disable=line-too-long
    {'fit': dict(inhibition_strength=0.1), 'nmf': dict(inhibition_range=((3, 3))), 'res': dict(energy=234.838968, norm_H_1=6730.89543, norm_H_0=176347)},  # noqa: E501
    {'fit': dict(inhibition_strength=0.5), 'nmf': dict(inhibition_range=((3, 3))), 'res': dict(energy=680.585424, norm_H_1=5177.87844, norm_H_0=174277)},  # noqa: E501
    {'fit': dict(inhibition_strength=1.0), 'nmf': dict(inhibition_range=((3, 3))), 'res': dict(energy=1119.00855, norm_H_1=4657.19574, norm_H_0=168777)},  # noqa: E501
    {'fit': dict(inhibition_strength=5.0), 'nmf': dict(inhibition_range=((3, 3))), 'res': dict(energy=518.936361, norm_H_1=6872.57858, norm_H_0=100488)},  # noqa: E501
    {'fit': dict(inhibition_strength=10.), 'nmf': dict(inhibition_range=((3, 3))), 'res': dict(energy=489.935256, norm_H_1=7224.76002, norm_H_0=62017)},  # noqa: E501
]

# create the input by concatenating the test image twice
img = racoon_image(gray=False, scale=0.1)
V = np.repeat(img.transpose((2, 0, 1))[np.newaxis, ...], 2, axis=0)


def _do_test(backend, params):
    # use the same random seed for all runs
    np.random.seed(seed=42)

    # create and fit the NMF model
    nmf = TransformInvariantNMF(
        n_atoms=10,
        atom_shape=(7, 7),
        n_iterations=25,
        backend=backend,
        verbose=3,
        **params['nmf']
    )
    nmf.fit(V, **params['fit'])

    H = nmf.H

    energy = nmf._energy_function(V)  # pylint: disable=protected-access
    norm_H_1 = np.sum(np.abs(H))
    norm_H_0 = np.sum(H/H.max() > 1e-7)

    nmf._logger.debug(f'energy={energy}, norm_H_1={norm_H_1}, norm_H_0={norm_H_0}')  # pylint: disable=protected-access

    expectation = params['res']
    assert np.isclose(energy, expectation['energy'])
    assert np.isclose(norm_H_1, expectation['norm_H_1'])
    assert np.isclose(norm_H_0, expectation['norm_H_0'])


@pytest.mark.parametrize('params', test_params_sparsity)
@pytest.mark.parametrize('backend', backends)
def test_sparsity(backend: str, params: Dict):
    _do_test(backend, params)


@pytest.mark.parametrize('params', test_params_inhibition)
@pytest.mark.parametrize('backend', backends)
def test_inhibition(backend: str, params: Dict):
    _do_test(backend, params)



"""
A module that provides some specializations and utilities for all NumPy based backends.
"""

from typing import Tuple, Optional

import numpy as np
from scipy.ndimage import convolve1d

from ._Backend import Backend


# pylint: disable=abstract-method
class NumPyBackend(Backend):
    r"""
    The parent class for all NumPy based backends.

    They provide the functionality to evaluate the analytic gradients of the factorization model.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._pad_mode = None
        self._padding_left = None
        self._padding_right = None

    def _initialize_matrices(
        self,
        V: np.ndarray,
        atom_shape: Tuple[int, ...],
        n_atoms: int,
        W: Optional[np.ndarray] = None,
    ) -> Tuple[np.ndarray, np.ndarray]:

        self._padding_left = tuple((s - 1, 0) for s in self.atom_shape)
        self._padding_right = tuple((0, s - 1) for s in self.atom_shape)

        if self._reconstruction_mode == 'valid':
            self._pad_mode = None
        elif self._reconstruction_mode == 'full':
            self._pad_mode = dict(mode='constant', constant_values=0.)
        elif self._reconstruction_mode == 'circular':
            self._pad_mode = dict(mode='wrap')
        elif self._reconstruction_mode == 'reflect':
            self._pad_mode = dict(mode='reflect', reflect_type='even')
        else:
            raise ValueError(f'Unsupported reconstruction mode "{self._reconstruction_mode}".'
                             f'Please choose "valid", "full", "circular", or "reflect".')

        return super()._initialize_matrices(V, atom_shape, n_atoms, W)

    @staticmethod
    def to_ndarray(arr: np.ndarray) -> np.ndarray:
        return arr

    @staticmethod
    def convolve_multi_1d(arr: np.ndarray, kernels: Tuple[np.ndarray, ...], axes: Tuple[int, ...]) -> np.ndarray:
        assert len(kernels) == len(axes)

        convolved = arr
        for a, kernel in zip(axes, kernels):
            convolved = convolve1d(convolved, kernel, axis=a, mode='constant', cval=0.0)

        return convolved



"""
Test the decomposition on multiple 1D curves and ensure that all backends yield the same factorization.
"""

import logging
from typing import Tuple

import numpy as np
import pytest

from tnmf.TransformInvariantNMF import TransformInvariantNMF


logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s: %(message)s')

# hard-coded expected energy levels for the different reconstruction modes
expected_energies = {
    'valid': 2.34946,
    'full': 1.87180,
    'circular': 3.13228,
    'reflect': 3.16430,
}

# define all test settings
backends = ['numpy', 'numpy_fft', 'numpy_caching_fft', 'pytorch', 'pytorch_fft']
reconstruction_modes = ['valid', 'full', 'circular']  # , 'reflect']

# temporarily ignore failed tests due to unimplemented features
raise_not_implemented_errors = False

# three periodic curves as input
V = np.array([[1., 2., 3., 2., 1., 1., 2., 3., 2., 1., 1., 2., 3., 2., 1.],
              [1., 2., 2., 2., 1., 1., 2., 2., 2., 1., 1., 2., 2., 2., 1.],
              [0., 1., 2., 3., 4., 0., 1., 2., 3., 4., 0., 1., 2., 3., 4.]])
# need a (singleton) channel axis
V = V[:, np.newaxis, :]


def fit_nmf(backend, reconstruction_mode):
    # use the same random seed for all runs
    np.random.seed(seed=42)

    # create and fit the NMF model
    nmf = TransformInvariantNMF(
        n_atoms=3,
        atom_shape=(5, ),
        n_iterations=10,
        backend=backend,
        verbose=3,
        reconstruction_mode=reconstruction_mode,
    )
    nmf.fit(V, inhibition_strength=0.1)

    return nmf


@pytest.fixture(name='expected_factorization')
def fixture_expected_factorization(reconstruction_mode):
    nmf = fit_nmf('numpy_fft', reconstruction_mode)
    return nmf.W, nmf.H


@pytest.mark.parametrize('reconstruction_mode', reconstruction_modes)
@pytest.mark.parametrize('backend', backends)
def test_expected_energy(backend: str, reconstruction_mode: str, expected_factorization: Tuple[np.ndarray, np.ndarray]):

    # extract the target tensors and the reconstruction mode dependent expected energy level
    W, H = expected_factorization
    expected_energy = expected_energies[reconstruction_mode]

    # fit the NMF model
    try:
        nmf = fit_nmf(backend, reconstruction_mode)
    except NotImplementedError as e:
        if raise_not_implemented_errors:
            raise AssertionError from e
        return

    nmf._logger.debug(f'W:\n{nmf.W}')  # pylint: disable=protected-access
    nmf._logger.debug(f'H:\n{nmf.H}')  # pylint: disable=protected-access
    nmf._logger.debug(f'R:\n{nmf.R}')  # pylint: disable=protected-access

    # check if the expected energy level is reached
    assert np.isclose(nmf._energy_function(V), expected_energy)  # pylint: disable=protected-access

    # check if the expected factorization is obtained
    assert np.allclose(nmf.W, W)
    assert np.allclose(nmf.H, H)

    # check if the atoms have unit norm
    norm_W = np.sum(nmf.W, axis=(-1,))
    assert np.allclose(norm_W, 1.)



import matplotlib.pyplot as plt
import numpy as np
import streamlit as st

from tnmf.TransformInvariantNMF import TransformInvariantNMF
from tnmf.utils.data_loading import racoon_image

progress_bar = st.sidebar.progress(0)

st.sidebar.markdown('# Input options')
scale = st.sidebar.slider('Image rescaling factor', min_value=0., max_value=1., value=0.2)
color = st.sidebar.checkbox('Color channels', False)

st.sidebar.markdown('# NMF options')
backend = st.sidebar.selectbox('Backend', ['numpy', 'numpy_fft', 'numpy_caching_fft', 'pytorch', 'pytorch_fft'], 1)
n_atoms = st.sidebar.number_input('Number of atoms', min_value=1, value=10)
atom_size = st.sidebar.number_input('Atom size', min_value=1, value=10)
sparsity_H = st.sidebar.number_input('Activation sparsity', min_value=0., value=0., step=0.1)
inhibition = st.sidebar.number_input('Activation lateral inhibition', min_value=0., value=0., step=0.1)
if inhibition > 0:
    inhibition_range = st.sidebar.selectbox('Inhibition Range', [f'default ({atom_size-1})', ] + list(range(1, atom_size)), 0)
    inhibition_range = None if isinstance(inhibition_range, str) else (inhibition_range, ) * 2
else:
    inhibition_range = None

n_iterations = st.sidebar.number_input('Number of iterations', min_value=1, value=100)

img = racoon_image(gray=not color, scale=scale)
V = img.transpose((2, 0, 1))[np.newaxis, ...] if color else img[np.newaxis, np.newaxis, ...]

nmf = TransformInvariantNMF(
    n_atoms=n_atoms,
    atom_shape=(atom_size, atom_size),
    n_iterations=n_iterations,
    backend=backend,
    inhibition_range=inhibition_range,
)

nmf.fit(
    V,
    sparsity_H=sparsity_H,
    inhibition_strength=inhibition,
    progress_callback=lambda _, x: progress_bar.progress((x + 1) / n_iterations)
)

img_cmap = None if color else 'gray'

st.markdown('# Input and reconstruction')
col1, col2 = st.beta_columns(2)
with col1:
    fig = plt.figure()
    plt.imshow(img, cmap=img_cmap)
    st.pyplot(fig)
with col2:
    fig = plt.figure()
    plt.imshow(nmf.R[0].transpose((1, 2, 0)), cmap=img_cmap)
    st.pyplot(fig)

st.markdown('## Atom, activation and partial reconstruction')
for i_atom in range(n_atoms):
    col1, col2, col3 = st.beta_columns(3)
    with col1:
        fig = plt.figure()
        plt.imshow(nmf.W[i_atom, 0], cmap=img_cmap)
        st.pyplot(fig)
    with col2:
        fig = plt.figure()
        plt.imshow(nmf.H[0, i_atom])
        st.pyplot(fig)
    with col3:
        fig = plt.figure()
        partial_R = nmf.R_partial(i_atom)
        plt.imshow(partial_R[0].transpose((1, 2, 0)) / partial_R[0].max(), cmap=img_cmap)
        st.pyplot(fig)



"""
A module that provides a NumPy based backend for computing the gradients of the factorization model.
Shift-invariance is implemented via fast convolution in the Fourier domain using :func:`scipy.fft.rfftn`
and :func:`scipy.fft.irfftn` with additional caching of the Fourier transformed arrays compared to
:mod:`tnmf.backends.NumPy_FFT`.
"""
# TODO: consider adding shape getters to CachingFFT
# TODO: this backend has a logger member but the other backends don't

import logging
from typing import Dict, Tuple, Optional, Union
from copy import copy

import numpy as np
from opt_einsum import contract_expression
from opt_einsum.contract import ContractExpression
from scipy.fft import next_fast_len, rfftn, irfftn
from scipy.ndimage import convolve1d

from ._NumPyFFTBackend import NumPyFFTBackend


class CachingFFT:
    """
    Wrapper class for conveniently caching and switching back and forth between arrays in coordinate space and their
    representations in Fourier space.
    """
    def __init__(
        self,
        field_name: str,
        fft_axes: Optional[Tuple[int, ...]] = None,
        fft_shape: Optional[Tuple[int, ...]] = None,
        logger: Optional[logging.Logger] = None,
    ):
        self._c = None  # field in coordinate space
        self._f = None  # field in fourier space
        self._f_padded = None  # fourier transform of padded field
        self._f_reversed = None  # time-reversed field in fourier space
        self._fft_axes = fft_axes
        self._fft_shape = fft_shape
        self._fft_workers = -1
        self._logger = logger if logger is not None else logging.getLogger(self.__class__.__name__)
        self._field_name = field_name

    def __imul__(self, other):
        self.c *= other.c if isinstance(other, CachingFFT) else other
        return self

    def __isub__(self, other):
        self.c -= other.c if isinstance(other, CachingFFT) else other
        return self

    def __iadd__(self, other):
        self.c += other.c if isinstance(other, CachingFFT) else other
        return self

    def __itruediv__(self, other):
        self.c /= other.c if isinstance(other, CachingFFT) else other
        return self

    def set_fft_params(self, fft_axes: Tuple[int, ...], fft_shape: Tuple[int, ...]):
        self._fft_axes = fft_axes
        self._fft_shape = fft_shape

    def _invalidate(self):
        self._c = None
        self._f = None
        self._f_padded = None
        self._f_reversed = None

    def has_c(self) -> bool:
        """Check if the field in coordinate space has already been computed"""
        return self._c is not None

    def has_f(self) -> bool:
        """Check if the field in fourier space has already been computed"""
        return self._f is not None

    @property
    def shape(self) -> Tuple[int]:
        return self._c.shape

    @property
    def ndim(self) -> int:
        return self._c.ndim

    @property
    def c(self) -> np.ndarray:
        """Getter for field in coordinate space"""
        if self._c is None:
            self._logger.debug(f'Computing {self._field_name}(x) = FFT^-1[ {self._field_name}(f) ]', )
            assert self.has_f
            self._c = irfftn(self._f, axes=self._fft_axes, s=self._fft_shape, workers=self._fft_workers)
        return self._c

    @c.setter
    def c(self, c: np.ndarray):
        """Setter for field in coordinate space"""
        self._logger.debug(f'{"Setting" if c is not None else "Clearing"} {self._field_name}(x)')
        self._invalidate()
        self._c = c

    @property
    def f(self) -> np.ndarray:
        """Getter for field in fourier space"""
        if self._f is None:
            self._logger.debug(f'Computing {self._field_name}(f) = FFT[ {self._field_name}(x) ]')
            assert self.has_c()
            self._f = rfftn(self._c, axes=self._fft_axes, s=self._fft_shape, workers=self._fft_workers)
        return self._f

    def f_padded(self, pad_mode: Dict = None, pad_width: Tuple[Tuple[int, int], ...] = None,) -> np.ndarray:
        """Getter for padded field in fourier space"""
        if self._f_padded is None:
            self._logger.debug(f'Computing {self._field_name}_padded(f) = FFT[ {self._field_name}_padded(x) ]')
            assert self.has_c()

            c = np.pad(self._c, pad_width, **pad_mode)
            # TODO: we should actually make sure that the padding does not change between calls
            self._f_padded = rfftn(c, axes=self._fft_axes, s=self._fft_shape, workers=self._fft_workers)
        return self._f_padded

    @f.setter
    def f(self, f: np.ndarray):
        """Setter for field in fourier space"""
        self._logger.debug(f'{"Setting" if f is not None else "Clearing"} {self._field_name}(f)')
        self._invalidate()
        self._f = f

    @property
    def f_reversed(self) -> np.ndarray:
        """Getter for time-reversed field in fourier space, intentionally no setter for now"""
        if self._f_reversed is None:
            self._logger.debug(f'Computing {self._field_name}_rev(f) = FFT[ {self._field_name}(-x) ]')
            assert self.has_c()
            c_reversed = np.flip(self._c, axis=self._fft_axes)
            self._f_reversed = rfftn(c_reversed, axes=self._fft_axes, s=self._fft_shape, workers=self._fft_workers)
        return self._f_reversed


class NumPy_CachingFFT_Backend(NumPyFFTBackend):
    r"""
    A NumPy based backend that performs convolutions and contractions for computing the gradients of the factorization model
    via FFT, similar to :class:`.NumPy_FFT_Backend`. However, the Fourier representations of the associated arrays are cached
    in order to reduce the number of Fourier transformations involved to a minimum.
    """
    def __init__(
        self,
        logger: logging.Logger = None,
        verbose: int = 0,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._logger = logger if logger is not None else logging.getLogger(self.__class__.__name__)
        self._logger.setLevel([logging.ERROR, logging.WARNING, logging.INFO, logging.DEBUG][verbose])
        self._V = None

    def _initialize_matrices(
        self,
        V: np.ndarray,
        atom_shape: Tuple[int, ...],
        n_atoms: int,
        W: Optional[np.ndarray] = None,
    ) -> Tuple[np.ndarray, np.ndarray]:

        w, h = super()._initialize_matrices(V, atom_shape, n_atoms, W)

        # fft shape and functions
        fft_axes = self._shift_dimensions
        fft_shape = np.array(self._sample_shape) + np.array(self._transform_shape) - 1
        if self._pad_mode is not None:
            fft_shape += np.asarray(self._padding_right).sum(axis=1)
        fft_shape = tuple(next_fast_len(s) for s in np.array(self._sample_shape) + np.array(self._transform_shape) - 1)

        self._V = CachingFFT('V', fft_axes=fft_axes, fft_shape=fft_shape, logger=self._logger)
        self._V.c = V

        H = CachingFFT('H', fft_axes=fft_axes, fft_shape=fft_shape, logger=self._logger)
        H.c = h

        if W is None:
            W = CachingFFT('W', fft_axes=fft_axes, fft_shape=fft_shape, logger=self._logger)
            W.c = w

        # add unpadded and unsliced axes and the necessary fft_shape param
        unpadded = ((0, 0), ) * (V.ndim - len(self._shift_dimensions))
        unsliced = (slice(None), ) * (V.ndim - len(self._shift_dimensions))
        for key in self.fft_params:
            self.fft_params[key]['pad_width'] = unpadded + self.fft_params[key]['pad_width']
            self.fft_params[key]['slices'] = unsliced + self.fft_params[key]['slices']
            self.fft_params[key]['fft_shape'] = fft_shape

        # sum_c V|R[n, c, ... ] * W[m , c, ...] --> dR / dH[n, m, ...]
        self.fft_params['reconstruct']['contraction'] = contract_expression(
            'nm...,mc...->nc...', H.f.shape, W.f.shape)

        # sum_c V|R[n, c, ... ] * W[m , c, ...] --> dR / dH[n, m, ...]
        self.fft_params['grad_H']['contraction'] = contract_expression(
            'nc...,mc...->nm...', self._V.f.shape, W.f_reversed.shape)

        # sum_n V|R[n, c, ... ] * H[n, m, ...]   --> dR / dW[m, c, ...]
        self.fft_params['grad_W']['contraction'] = contract_expression(  # TODO: why is the pylint annotation necessary?
            'nc...,nm...->mc...', self._V.f.shape, H.f_reversed.shape)   # pylint: disable=no-member

        return W, H

    @staticmethod
    def to_ndarray(arr: CachingFFT) -> np.ndarray:
        return arr.c

    @staticmethod
    def normalize(arr: CachingFFT, axis: Optional[Union[int, Tuple[int, ...]]] = None):
        # TODO: overwriting the parent method can be avoided by redefining the division operator of CachingFFT and defining
        #   a common "array type" that can handle both np.ndarray and CachingFFT objects
        arr.c /= arr.c.sum(axis=axis, keepdims=True)

    @staticmethod
    def convolve_multi_1d(arr: CachingFFT, kernels: Tuple[np.ndarray, ...], axes: Tuple[int, ...]) -> CachingFFT:
        assert len(kernels) == len(axes)

        convolved = copy(arr)
        for a, kernel in zip(axes, kernels):
            # TODO: it should be possible to formulate this in Fourier space
            convolved.c = convolve1d(convolved.c, kernel, axis=a, mode='constant', cval=0.0)

        return convolved

    @staticmethod
    def _fft_convolve(
        name: Tuple[str, ...],
        arr1: Tuple[np.ndarray, ...],
        arr2: np.ndarray,
        contraction: ContractExpression,
        slices: Tuple[slice, ...],
        fft_axes: Tuple[int, ...],
        fft_shape: Tuple[int, ...],
        pad_mode: Dict = None,
        pad_width: Tuple[Tuple[int, int], ...] = None,
        correlate: bool = False,
        arr1_slice: Tuple[slice, ...] = None,
        arr2_slice: Tuple[slice, ...] = None,
    ) -> CachingFFT:
        arr1_fft = (a.f if pad_mode is None else a.f_padded(pad_mode, pad_width) for a in arr1)
        arr2_fft = arr2.f_reversed if correlate else arr2.f

        arr1_slice = slice(None) if arr1_slice is None else arr1_slice
        arr2_slice = slice(None) if arr2_slice is None else arr2_slice

        ret = tuple()
        for a, n in zip(arr1_fft, name):
            result = CachingFFT(n, fft_axes=fft_axes, fft_shape=fft_shape)
            result.f = contraction(a[arr1_slice], arr2_fft[arr2_slice])
            result.c = result.c[slices]
            ret += (result, )
        return ret

    def reconstruction_gradient_W(self, V: np.ndarray, W: CachingFFT, H: CachingFFT) -> Tuple[CachingFFT, CachingFFT]:
        R = self.reconstruct(W, H)
        assert R.c.shape == V.shape
        return self._fft_convolve(('neg_W', 'pos_W'), (self._V, R), H, **self.fft_params['grad_W'])

    def reconstruction_gradient_H(self, V: np.ndarray, W: CachingFFT, H: CachingFFT) -> Tuple[CachingFFT, CachingFFT]:
        R = self.reconstruct(W, H)
        assert R.c.shape == V.shape
        return self._fft_convolve(('neg_H', 'pos_H'), (self._V, R), W, **self.fft_params['grad_H'])

    def reconstruct(self, W: CachingFFT, H: CachingFFT) -> CachingFFT:
        R, = self._fft_convolve(('R', ), (H, ), W, **self.fft_params['reconstruct'])
        return R

    def partial_reconstruct(self, W: np.ndarray, H: np.ndarray, i_atom: int) -> np.ndarray:
        R_partial, = self._fft_convolve(
            ('R_partial', ), (H, ), W, **self.fft_params['reconstruct'],
            arr1_slice=(slice(None), slice(i_atom, i_atom+1)),
            arr2_slice=(slice(i_atom, i_atom+1)),
            )
        return R_partial



"""
A module that defines a common interface for all backends.
"""

# TODO: generalize n_transforms from numpy_fft to all backends
# TODO: all backends need to support self._mode_R et al
# TODO: do we need self._input_padding ? If yes, all backends have to support it.
# TODO: refactor common backend logic of NumpyBackend/PyTorchBackend into function

from abc import ABC, abstractmethod
from typing import Tuple, Optional, Union

import numpy as np


class Backend(ABC):
    r"""
    The base class for all backends.

    It exists to define a generic functional interface that is shared by all backends through inheritance, which allows a
    clear separation between algorithmic steps and their actual numerical implementation.

    Some common functionality between all backends is already defined inside this class but can be overridden if necessary.
    """
    def __init__(
        self,
        reconstruction_mode: str = 'valid',
    ):
        self._reconstruction_mode = reconstruction_mode
        self.atom_shape = None
        self.n_samples = None
        self.n_channels = None
        self._sample_shape = None
        self._transform_shape = None
        self._n_shift_dimensions = None
        self._shift_dimensions = None

    def initialize(
        self,
        V: np.ndarray,
        atom_shape: Tuple[int, ...],
        n_atoms: int,
        W: Optional[np.ndarray] = None,
    ) -> Tuple[np.ndarray, np.ndarray]:
        self._set_dimensions(V, atom_shape)
        return self._initialize_matrices(V, atom_shape, n_atoms, W)

    @staticmethod
    @abstractmethod
    def to_ndarray(arr) -> np.ndarray:
        raise NotImplementedError

    def _set_dimensions(self, V, atom_shape):
        self.atom_shape = atom_shape
        self.n_samples = V.shape[0]
        self.n_channels = V.shape[1]
        self._sample_shape = V.shape[2:]
        self._transform_shape = self._n_transforms(self._reconstruction_mode, self._sample_shape, atom_shape)
        self._n_shift_dimensions = len(atom_shape)
        self._shift_dimensions = tuple(range(-1, -len(atom_shape) - 1, -1))

    @staticmethod
    def _n_transforms(reconstruction_mode: str, sample_shape: Tuple[int, ...], atom_shape: Tuple[int, ...]) -> Tuple[int, ...]:
        # TODO: remove or rename this function
        """Number of dictionary transforms in each dimension"""
        if reconstruction_mode == 'valid':
            return tuple(np.array(sample_shape) + np.array(atom_shape) - 1)

        if reconstruction_mode == 'full':
            return tuple(np.array(sample_shape) - np.array(atom_shape) + 1)

        if reconstruction_mode in ('same', 'circular', 'reflect'):
            return tuple(np.array(sample_shape))

        raise ValueError

    @staticmethod
    def normalize(arr: np.ndarray, axis: Optional[Union[int, Tuple[int, ...]]] = None):
        arr /= (arr.sum(axis=axis, keepdims=True))

    @staticmethod
    def convolve_multi_1d(arr: np.ndarray, kernels: Tuple[np.ndarray, ...], axes: Tuple[int, ...]) -> np.ndarray:
        raise NotImplementedError

    def _initialize_matrices(
        self,
        V: np.ndarray,
        atom_shape: Tuple[int, ...],
        n_atoms: int,
        W: Optional[np.ndarray] = None,
    ) -> Tuple[np.ndarray, np.ndarray]:

        H = np.asarray(1 - np.random.rand(self.n_samples, n_atoms, *self._transform_shape), dtype=V.dtype)

        if W is None:
            W = np.asarray(1 - np.random.rand(n_atoms, self.n_channels, *atom_shape), dtype=V.dtype)

        return W, H

    @abstractmethod
    def reconstruction_gradient_W(self, V: np.ndarray, W: np.ndarray, H: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        raise NotImplementedError

    @abstractmethod
    def reconstruction_gradient_H(self, V: np.ndarray, W: np.ndarray, H: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        raise NotImplementedError

    @abstractmethod
    def reconstruct(self, W: np.ndarray, H: np.ndarray) -> np.ndarray:
        raise NotImplementedError

    def partial_reconstruct(self, W: np.ndarray, H: np.ndarray, i_atom: int) -> np.ndarray:
        return self.reconstruct(W[i_atom:i_atom+1], H[:, i_atom:i_atom+1])

    def reconstruction_energy(self, V: np.ndarray, W: np.ndarray, H: np.ndarray) -> float:
        R = self.to_ndarray(self.reconstruct(W, H))
        assert R.shape == V.shape
        return 0.5 * np.sum(np.square(V - R))



"""
A module that provides a PyTorch based backend for computing the gradients of the factorization model.
Shift-invariance is implemented via explicit convolution operations in the coordinate space.
"""

# TODO: use torch.fft.rfftn() to generalize for more dimensions and improve performance

import torch
from torch import Tensor
from torch.nn.functional import pad

from ._PyTorchBackend import PyTorchBackend


#: Lookup table for the convolution function with different dimensionality
_CONV_DICT = {
    1: torch.nn.functional.conv1d,
    2: torch.nn.functional.conv2d,
    3: torch.nn.functional.conv3d,
}


class PyTorch_Backend(PyTorchBackend):
    r"""
    Reconstruction is performed via an explicit convolution in coordinate space.
    """

    def reconstruct(self, W: Tensor, H: Tensor) -> Tensor:
        # TODO: support dimensions > 3
        # TODO: consider transposed convolution as alternative

        n_shift_dimensions = W.ndim - 2

        assert n_shift_dimensions <= 3
        conv_fun = _CONV_DICT[n_shift_dimensions]
        flip_dims = list(range(-n_shift_dimensions, 0))
        W_flipped = torch.flip(W, flip_dims)

        if self._padding is None:
            H_padded = H
        else:
            H_padded = pad(H, **self._padding)

        R = conv_fun(H_padded, torch.swapaxes(W_flipped, 0, 1))
        return R



import streamlit as st
from streamlit.delta_generator import DeltaGenerator

from tnmf.TransformInvariantNMF import TransformInvariantNMF
from tnmf.utils.demo import SignalTool, st_define_nmf_params


@st.cache(hash_funcs={DeltaGenerator: lambda _: None})
def fit_nmf_model(V, nmf_params, progress_bar):
    fit_params = {k: nmf_params.pop(k) for k in ('sparsity_H', 'inhibition_strength')}
    nmf = TransformInvariantNMF(**nmf_params)
    nmf.fit(V, progress_callback=lambda _, x: progress_bar.progress((x + 1) / nmf_params['n_iterations']), **fit_params)
    return nmf


def main(progress_bar, n_dims: int, verbose: bool = True):
    """
    Runs the streamlit demo on synthetic signals for the specified signal type.

    Parameters
    ----------
    progress_bar
        Streamlit progress bar that needs to be updated during model fitting.
    n_dims : int
        The number of dimensions of the input signals.
    verbose : bool
        If True, show detailed information.
    """
    # show demo info text
    if verbose:
        st.markdown('''
        ## Demo guide
        This dashboard demonstrates the use of the *Transform-Invariant Non-Negative Matrix Factorization (TNMF) package*
        in the specific context of learning **shift-invariant representations**.
        Via the **drop-down menu** on the left, you can choose between different demo examples.

        ## Usage
        Detailed explanations of the available options and the shown output (including this text) can be displayed or hidden
        using the **'Verbose' checkbox**. When the checkbox is unticked, the description of the control widgets can still
        be accessed via the **tooltips** next to them.
        ''')

    # create the signal handling tool
    tool = SignalTool(n_dims=n_dims)

    # generate random input data for the factorization
    V, opt_nmf_params = tool.st_generate_input(verbose=verbose)

    # define the NMF parameters and fit the model
    nmf_params = st_define_nmf_params(opt_nmf_params, verbose=verbose)
    nmf = fit_nmf_model(V, nmf_params, progress_bar)

    # visualize the results
    tool.st_compare_signals(V, nmf.R, verbose=verbose)
    tool.st_compare_individual_signals(V, nmf.R, verbose=verbose)
    tool.st_plot_partial_reconstructions(V, nmf, verbose=verbose)



"""
A module that provides a NumPy based backend for computing the gradients of the factorization model.
Shift-invariance is implemented via explicit convolution operations in the coordinate space.
"""
from itertools import product
from typing import Tuple, Optional

import numpy as np
from numpy.lib.stride_tricks import as_strided
from opt_einsum import contract

from ._NumPyBackend import NumPyBackend


class NumPy_Backend(NumPyBackend):
    r"""
    A plain NumPy backend for computing the gradients of the factorization model in coordinate space (no FFT, no PyTorch).

    Convolutions are computed efficiently as contractions of properly strided arrays.
    """
    def __init__(
        self,
        reconstruction_mode: str = 'valid',
    ):
        if reconstruction_mode != 'valid':
            raise NotImplementedError('This backend only supports the "valid" reconstruction mode.')
        super().__init__(reconstruction_mode=reconstruction_mode)
        self._shift_axes = None
        self._cache = {}

    def _initialize_matrices(
        self,
        V: np.ndarray,
        atom_shape: Tuple[int, ...],
        n_atoms: int,
        W: Optional[np.ndarray] = None,
    ) -> Tuple[np.ndarray, np.ndarray]:

        W, H = super()._initialize_matrices(V, atom_shape, n_atoms, W)

        n_shift_axes = len(self._sample_shape)
        self._shift_axes = tuple(range(-n_shift_axes, 0))
        shift_dim_idx = tuple(range(n_shift_axes))

        self._cache = {
            # zero-padding of the signal matrix for full-size correlation
            'pad_width': ((0, 0), (0, 0)) + tuple((a - 1, a - 1) for a in atom_shape)
        }

        self._cache.update({
            'V_padded': np.pad(V, pad_width=self._cache['pad_width']),
            # dimension labels of the data and reconstruction matrices
            'V_labels': ['n', 'c'] + ['d' + str(i) for i in shift_dim_idx],
            'W_labels': ['m', 'c'] + ['a' + str(i) for i in shift_dim_idx],
            'H_labels': ['n', 'm'] + ['d' + str(i) for i in shift_dim_idx],
            # dimension info for striding in gradient_H computation
            'X_strided_W_shape': V.shape[:2] + H.shape[2:] + atom_shape,
            'X_strided_W_labels': ['n', 'c'] + [s + str(i) for s, i in product(['d', 'a'], shift_dim_idx)],
            # dimension info for striding in gradient_W computation
            'H_strided_V_shape': H.shape[:2] + atom_shape + V.shape[2:],
            'H_strided_V_labels': ['n', 'm'] + [s + str(i) for s, i in product(['a', 'd'], shift_dim_idx)],
            # dimension info for striding in reconstruction computation
            'H_strided_W_shape': H.shape[:2] + V.shape[2:] + atom_shape,
            'H_strided_W_labels': ['n', 'm'] + [s + str(i) for s, i in product(['d', 'a'], shift_dim_idx)],
        })

        return W, H

    def reconstruction_gradient_W(self, V: np.ndarray, W: np.ndarray, H: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        H_strided_V_strides = H.strides + H.strides[2:]  # do NOT put these strides into self._cache as H layout can change
        H_strided = as_strided(H, self._cache['H_strided_V_shape'], H_strided_V_strides, writeable=False)
        R = self.reconstruct(W, H)

        neg = np.flip(contract(
            H_strided, self._cache['H_strided_V_labels'],
            V, self._cache['V_labels'],
            self._cache['W_labels'], optimize='optimal'), axis=self._shift_axes)

        pos = np.flip(contract(
            H_strided, self._cache['H_strided_V_labels'],
            R, self._cache['V_labels'],
            self._cache['W_labels'], optimize='optimal'), axis=self._shift_axes)
        return neg, pos

    def reconstruction_gradient_H(self, V: np.ndarray, W: np.ndarray, H: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        V_padded = self._cache['V_padded']
        V_strided_W_strides = V_padded.strides + V_padded.strides[2:]  # do NOT put these strides into self._cache
        V_strided = as_strided(V_padded, self._cache['X_strided_W_shape'], V_strided_W_strides, writeable=False)
        neg = contract(
            W, self._cache['W_labels'],
            V_strided, self._cache['X_strided_W_labels'],
            self._cache['H_labels'], optimize='optimal')

        R_padded = np.pad(self.reconstruct(W, H), pad_width=self._cache['pad_width'])
        R_strided_W_strides = R_padded.strides + R_padded.strides[2:]  # do NOT put these strides into self._cache
        R_strided = as_strided(R_padded, self._cache['X_strided_W_shape'], R_strided_W_strides, writeable=False)
        pos = contract(
            W, self._cache['W_labels'],
            R_strided, self._cache['X_strided_W_labels'],
            self._cache['H_labels'], optimize='optimal')
        return neg, pos

    def reconstruct(self, W: np.ndarray, H: np.ndarray) -> np.ndarray:
        H_strided_W_strides = H.strides + H.strides[2:]  # do NOT put these strides into self._cache as H layout can change
        H_strided = as_strided(H, self._cache['H_strided_W_shape'], H_strided_W_strides, writeable=False)
        R = contract(
            H_strided, self._cache['H_strided_W_labels'],
            np.flip(W, self._shift_axes), self._cache['W_labels'],
            self._cache['V_labels'], optimize='optimal')
        return R



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
sys.path.insert(0, os.path.abspath('..'))


# -- Project information -----------------------------------------------------

project = 'TransformInvariantNMF'
copyright = '2021, Adrian Šošić, Mathias Winkel'
author = 'Adrian Šošić, Mathias Winkel'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'autoapi.extension',
    'sphinx.ext.intersphinx',
    'sphinx.ext.napoleon',    # https://www.sphinx-doc.org/en/master/usage/extensions/napoleon.html#module-sphinx.ext.napoleon
    'm2r2',
]

autoapi_type = 'python'
autoapi_dirs = ['../tnmf', ]
autoapi_options = ['members', 'undoc-members', 'no-private-members', 'show-inheritance',
                   'show-module-summary', 'special-members', 'imported-members', ]

napoleon_attr_annotations = True
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = False

m2r_parse_relative_links = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static', ]
html_logo = "../logos/tnmf_logo.svg"
html_theme_options = {
    'logo_only': False,
    'display_version': True,
}

html_css_files = [
    'css/emdgroup.css',
]

intersphinx_mapping = {
    'python': ('https://docs.python.org/{.major}'.format(sys.version_info), None),
    'numpy': ('https://numpy.org/doc/stable', None),
    'sklearn': ('https://scikit-learn.org/stable', None),
    'scipy': ('https://docs.scipy.org/doc/scipy/reference', None),
    "torch": ("https://pytorch.org/docs/master/", None),
    }



"""
A module that provides a PyTorch based backend for computing the gradients of the factorization model.
Shift-invariance is implemented via efficient convolution operations in Fourier space.
"""

import torch
from torch import Tensor
from torch.nn.functional import pad
from torch.fft import rfftn, irfftn

from ._PyTorchBackend import PyTorchBackend


class PyTorch_FFT_Backend(PyTorchBackend):
    r"""
    Reconstruction is performed via an efficient convolution in Fourier space.
    """

    def reconstruct(self, W: Tensor, H: Tensor) -> Tensor:
        """
        Compute sum_m    H[n, m, _, ... ] * W[_ , m, c, ...]   --> R[n, c, ...]
        """

        n_shift_dimensions = W.ndim - 2

        if self._padding is None:
            H_padded = H
        else:
            H_padded = pad(H, **self._padding)

        shape = H_padded.shape[-n_shift_dimensions:]
        fft_shape = shape

        # if dim is not specified, the last len(s) dimensions are transformed, which is exactly what we need
        W_fft = rfftn(W, s=fft_shape)
        H_fft = rfftn(H_padded, s=fft_shape)

        # insert singleton dimensions to ensure broadcasting works as required
        W_fft = torch.unsqueeze(W_fft, 0)
        H_fft = torch.unsqueeze(H_fft, 2)

        R_fft = W_fft * H_fft

        R_fft = R_fft.sum(dim=1, keepdim=False)

        fslice = (slice(None), ) * 2
        fslice += tuple(slice(min(H_padded.shape[a], W.shape[a]) - 1,
                              max(H_padded.shape[a], W.shape[a])) for a in range(-n_shift_dimensions, 0))

        R = irfftn(R_fft, s=shape)

        return R[fslice]



import argparse
import os
from glob import glob
from pathlib import Path
from typing import List

DEMO_FILE = Path(__file__).absolute().parent.parent / 'demos' / 'demo_selector.py'
EXAMPLES_FOLDER = Path(__file__).absolute().parent.parent / 'examples'


def get_examples() -> List[Path]:
    """Returns a list containing the absolute file paths of all examples."""
    file_list = glob(str(EXAMPLES_FOLDER / '*.py'))
    examples = []
    for p in file_list:
        abs_file_path = Path(p)
        if abs_file_path.name != '__init__.py':
            examples.append(abs_file_path)
    return examples


def main():
    """Entry point for the `tnmf` command."""

    # create parser for the `tnmf` command
    parser = argparse.ArgumentParser()

    # create subparsers for the `example` and `demo` commands
    subparsers = parser.add_subparsers(dest='command')
    demo_parser = subparsers.add_parser("demo", help='Runs the streamlit demo.')  # noqa: F841
    example_parser = subparsers.add_parser("example", help='Runs a specific example. Type `tnmf example -h` to show the list '
                                                           'of available examples.')

    examples = get_examples()
    example_parser.add_argument("example_name", choices=[e.stem for e in examples], help='Name of the example.')

    # parse the command type
    args = parser.parse_args()

    # parse the example name and run the corresponding script
    if args.command == 'example':
        args_example = parser.parse_args()
        example = [e for e in examples if e.stem == args_example.example_name][0]
        os.system(f'python {example}')  # noqa: S605

    # run the demo
    elif args.command == 'demo':
        os.system(f'streamlit run {DEMO_FILE}')  # noqa: S605



"""
A module that provides a NumPy based backend for computing the gradients of the factorization model.
Shift-invariance is implemented via fast convolution in the Fourier domain using :func:`scipy.fft.rfftn`
and :func:`scipy.fft.irfftn`.
"""
from typing import Tuple, Dict, Union

import numpy as np
from scipy.fft import next_fast_len, rfftn, irfftn

from ._NumPyFFTBackend import NumPyFFTBackend


def fftconvolve_sum(
        in1: Union[np.ndarray, Tuple[np.ndarray, ...]],
        in2: np.ndarray,
        fft_axes: Tuple[int, ...],
        slices: Tuple[slice, ...],
        correlate: bool,
        pad_mode: Dict = None,
        pad_width: Tuple[Tuple[int, int], ...] = None,
        sum_axis: Tuple[int, ...] = None,
        keepdims: bool = False,
) -> np.ndarray:

    def padded_rfftn(
        array: np.ndarray,
        fft_shape: Tuple[int, ...],
        axes: Tuple[int, ...],
        pad_mode: Dict = None,
        pad_shape: Dict = None
    ) -> np.ndarray:
        # padding to fulfill boundary conditions
        if pad_mode is not None:
            array = np.pad(array, pad_shape, **pad_mode)

        # first axes are not to be padded
        fftpadding = tuple(((0, 0), ) * (array.ndim - len(axes)))
        # last axes will be padded
        fftpadding += tuple((0, fft_shape[ia] - array.shape[a]) for ia, a in enumerate(axes))
        # zero-pad the relevant axes
        array_padded = np.pad(array, fftpadding, mode='constant', constant_values=0.)
        # Fourier transform (for real data)
        return rfftn(array_padded, axes=axes)

    if not isinstance(in1, tuple):
        assert isinstance(in1, np.ndarray)
        in1 = (in1, )

    assert isinstance(in2, np.ndarray)

    ndim = in2.ndim
    s1 = in1[0].shape
    s2 = in2.shape

    assert all(a.ndim == ndim for a in in1)
    assert all(a.shape == s1 for a in in1)

    # convert negative axes indices into positive counterparts
    axes = sorted([a if a >= 0 else a + ndim for a in fft_axes])

    # pad according to required boundary conditions
    if pad_mode is not None:
        # we need padding information for all axes to be convolved over
        assert pad_width is not None and len(pad_width) == len(axes)

        pad_shape = tuple(((0, 0), ) * (ndim - len(axes))) + pad_width
        padded_shape = np.asarray(pad_shape).sum(axis=1) + np.asarray(s1)
    else:
        pad_shape = None
        padded_shape = s1

    # now need to zero-pad to identical size / optimal fft size
    fft_shape = tuple(next_fast_len(padded_shape[a] + s2[a] - 1, True) for a in axes)

    # compute how to undo the padding
    fslice = (slice(None), ) * (ndim - len(axes)) + slices

    # for correlation instead of convolution, we simply reverse the second array
    if correlate:
        reverse = ((slice(None), )) * (ndim - len(fft_axes)) + (slice(None, None, -1),) * len(fft_axes)
        assert len(reverse) == len(s2)
        in2 = in2[reverse]

    # Fourier transform (for real data)
    sp1 = (padded_rfftn(in1i, fft_shape, axes, pad_mode, pad_shape) for in1i in in1)
    sp2 = padded_rfftn(in2, fft_shape, axes)

    result = tuple()

    for sp1i in sp1:
        # perform convolution by multiplication in Fourier space
        sp1sp2 = sp1i * sp2

        # sum over an axis if requested, set the corresponding dimension to 1
        if sum_axis is not None:
            sp1sp2 = np.sum(sp1sp2, axis=sum_axis, keepdims=True)

        # transform back to real space
        ret = np.asarray(irfftn(sp1sp2, axes=axes))

        # actually remove the padded rows/columns
        ret = ret[fslice].copy()

        # remove singleton dimensions if requested
        if not keepdims and sum_axis is not None:
            assert ret.shape[sum_axis] == 1
            ret = np.squeeze(ret, sum_axis)

        result += (ret, )

    return result


class NumPy_FFT_Backend(NumPyFFTBackend):
    r"""
    A NumPy based backend that performs convolutions and contractions for computing the gradients of the factorization model
    via FFT:

    Arrays to be convolved are transformed to Fourier space, multiplied and accumulated across the free indices (e.g. for the
    sum over all atoms in the reconstruction), and transformed back to coordinate space.
    """

    def reconstruction_gradient_W(self, V: np.ndarray, W: np.ndarray, H: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        R = self.reconstruct(W, H)
        assert R.shape == V.shape

        #          sum_n    H[n , m, _, ...] * V|R[n, _, c, ... ]   --> dR / dW[m, c, ...]
        neg, pos = fftconvolve_sum(
            (V[:, np.newaxis, :, ...], R[:, np.newaxis, :, ...]),
            H[:, :, np.newaxis, ...],
            sum_axis=0,
            **self.fft_params['grad_W'])

        assert neg.shape == W.shape
        assert pos.shape == W.shape

        return neg, pos

    def reconstruction_gradient_H(self, V: np.ndarray, W: np.ndarray, H: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        R = self.reconstruct(W, H)
        assert R.shape == V.shape

        #        sum_c        V|R[n, _, c, ... ] * W[_ , m, c, ...]   --> dR / dH[n, m, ...]
        neg, pos = fftconvolve_sum(
            (V[:, np.newaxis, :, ...], R[:, np.newaxis, :, ...]),
            W[np.newaxis, :, :, ...],
            sum_axis=2,
            **self.fft_params['grad_H'])

        assert neg.shape == H.shape
        assert pos.shape == H.shape

        return neg, pos

    def reconstruct(self, W: np.ndarray, H: np.ndarray) -> np.ndarray:
        #        sum_m    H[n, m, _, ... ] * W[_ , m, c, ...]   --> R[n, c, ...]
        R, = fftconvolve_sum(
            (H[:, :, np.newaxis, ...], ),
            W[np.newaxis, :, ...],
            sum_axis=1,
            **self.fft_params['reconstruct'])

        return R



"""Provides functionality to generate certain 1-D and 2-D test signals."""

# pylint: disable=redefined-outer-name, unnecessary-lambda

from itertools import product, zip_longest
from typing import Optional, List, Tuple

import matplotlib.pyplot as plt
import numpy as np
from more_itertools import chunked
from numpy.linalg import norm


def generate_pulse(shape: str, length: int = 20) -> np.ndarray:
    """
    Generates a short signal pulse of specified shape and length.

    Parameters
    ----------
    shape : 'n' | '-' | '^' | 'v' | '_'
        Shape of the pulse.
    length : int
        Length of the pulse.

    Returns
    -------
    pulse : np.ndarray
        The signal pulse as a 1-D array.
    """
    # used for ramp-shaped pulses
    l1, l2 = np.ceil(length / 2), np.floor(length / 2)

    # define the signals shape
    if shape == 'n':
        r = (length - 1) / 2
        f = lambda x: np.sqrt((r ** 2) - ((x - r) ** 2))  # noqa: E731

    elif shape == '-':
        f = lambda x: np.ones_like(x, dtype=float)  # noqa: E731

    elif shape == '^':
        f = lambda x: np.hstack([np.arange(l1), l1 - 1 - (l1 != l2) - np.arange(l2)])  # noqa: E731

    elif shape == 'v':
        f = lambda x: np.hstack([l1 - 1 - np.arange(l2), np.arange(l1)])  # noqa: E731

    elif shape == '_':
        f = lambda x: np.zeros_like(x)  # noqa: E731

    else:
        raise ValueError('unknown pulse shape')

    # create the pulse
    x = np.arange(length)
    pulse = f(x)

    # normalize the signal
    if shape != '_':
        pulse /= norm(pulse)

    return pulse


def generate_pulse_train(
        symbols: Optional[List[str]] = None,
        pulse_length: int = 20,
        n_pulses: int = 5,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Generates a signal composed of a random sequence of multi-channel pulses.

    Parameters
    ----------
    symbols : List[str], optional
        A list of symbols (= multi-channel pulse shapes) specified via strings that are used as a dictionary to generate
        the signal. If 'None', the following three-channel symbols are used: ['nnn', '---', '^^^', 'vvv', '___'].
        See `generate_pulse` for an overview of all available pulse shapes.
    pulse_length : int
        The length of each individual pulse.
    n_pulses : int
        The total number of pulses to be sequenced.

    Returns
    -------
    signal : np.ndarray
        A 2-D array of shape (n_channels, n_pulses * pulse_length) containing the signal.
    W : np.ndarray
        A 3-D array of shape (n_symbols, n_channels, pulse_length) representing the pulse shape dictionary.
    """
    # default pulse shapes
    if symbols is None:
        symbols = ['nnn', '---', '^^^', 'vvv', '___']

    # assert that all provided symbols have the same number of channels
    else:
        if len(np.unique([len(p) for p in symbols])) != 1:
            raise ValueError('all symbols must have the same number of channels')

    # generate all pulse shapes
    W = np.stack([np.stack([generate_pulse(shape, pulse_length) for shape in symbol]) for symbol in symbols])

    # generate a random sequence of pulse indices and synthesize the signal
    sequence = np.random.choice(range(len(symbols)), n_pulses)
    signal = np.hstack([W[symbol_idx] for symbol_idx in sequence])

    return signal, W


def generate_patch(pattern: str, size: int = 10, color: Optional[str] = None) -> np.ndarray:
    """
    Generates a square image patch showing a certain pattern of specified size and color.

    Parameters
    ----------
    pattern : 'x' | '+' | 's'
        Pattern shown in the image patch.
    size : int
        Size of both dimensions of the image patch.
    color : 'r' | 'g' | 'b' | 'y' | 'm' | 'c' | 'w' , optional
        Color of the pattern. If 'None', the generated image patch will be grayscale.

    Returns
    -------
    patch : np.ndarray
        The image patch as a 3-D array, where the first dimension indexes the color channel.
    """
    # generate the grayscale version of the image patch
    if pattern == 'x':
        im = np.eye(size)
        im[np.rot90(im).astype(bool)] = 1
    elif pattern == '+':
        im = np.zeros([size, size])
        idx = np.array([np.floor((size-1)/2), np.ceil((size-1)/2)]).astype(int)
        im[idx, :] = 1
        im[:, idx] = 1
    elif pattern == 's':
        fill_width = int(size/3)
        square_width = size - 2 * fill_width
        im = np.pad(np.ones([square_width, square_width]), fill_width)
    else:
        raise ValueError('unknown patch shape')

    # convert the grayscale image to RGB
    if color:
        patch = np.zeros([3, size, size])
        color_dict = {'r': [0], 'g': [1], 'b': [2], 'y': [0, 1], 'm': [0, 2], 'c': [1, 2], 'w': [0, 1, 2]}
        channels = color_dict[color]
        patch[channels] = np.tile(im[None], [len(channels), 1, 1])
    else:
        patch = im[None]

    return patch


def generate_block_image(
        symbols: Optional[List[str]] = None,
        symbol_size: int = 10,
        n_symbols: int = 10,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Generates a block-structured image composed of several image patches, each containing a single random pattern.

    Parameters
    ----------
    symbols : List[str], optional
        A list of symbols used as a dictionary to generate the image. Each symbol is either a one-character or
        two-character string, where the first character specifies the symbol shape and the optional second character
        specifies the symbol color. For example: 'x' creates a grayscale cross-shaped patch while 'sr' creates a red square-
        shaped patch. If 'None', a certain default set of colored patches will be used as dictionary. See `generate_patch`
        for an overview of all available patch shapes and color options.
    symbol_size : int
        The size of both dimensions of each individual image patch.
    n_symbols : int
        The number of image patches to be stacked both horizontally and vertically.

    Returns
    -------
    image : np.ndarray
        A 3-D array of shape (3, n_symbols * symbol_size, n_symbols * symbol_size) containing the image.
    W : np.ndarray
        A 3-D array of shape (3, symbols_size, symbol_size) containing the image patch dictionary.
    """
    # default symbols
    if symbols is None:
        shapes = ['+', 'x', 's']
        colors = ['r', 'g', 'b', 'y', 'm', 'c', 'w']
        symbols = [''.join(spec) for spec in product(shapes, colors)]

    # separate shape and color information if colors are specified
    if len(symbols[0]) == 1:
        symbols = list(symbols).copy()
        symbols[0] = (symbols[0], None)
    shapes, colors = zip_longest(*symbols)

    # generate a random sequence of patches indices
    sequence = np.random.choice(range(len(symbols)), n_symbols ** 2)

    # generate all patch types
    W = np.stack([generate_patch(shape, symbol_size, color) for shape, color in zip(shapes, colors)])

    # turn the sequence of patch indices into a sequence patches and stack them into an image
    patches = [W[patch_idx] for patch_idx in sequence]
    image = np.block(list(chunked(patches, n_symbols)))

    return image, W


if __name__ == '__main__':

    plt.style.use('seaborn')

    # ---------- 1-D example ---------- #

    # specify pulse properties
    n_pulses = 6
    pulse_length = 100

    # generate the pulse signal
    signal, _ = generate_pulse_train(pulse_length=pulse_length, n_pulses=n_pulses)

    # visualize the signal and highlight the individual pulses
    fig, axs = plt.subplots(nrows=signal.shape[1])
    for channel, ax in enumerate(axs):
        for p in range(n_pulses):
            x = range(p * pulse_length, p * pulse_length + pulse_length)
            ax.plot(x, signal[x, channel])
    plt.show()

    # ---------- 2-D example ---------- #

    # specify image patch properties
    symbol_size = 11
    n_symbols = 10

    # generate and visualize the image
    im, _ = generate_block_image(symbol_size=symbol_size, n_symbols=n_symbols)
    plt.imshow(im)
    plt.axis('off')
    plt.grid(False)
    plt.tight_layout()
    plt.show()



"""
=================================================================
Decomposition of a gray scale image into characteristic patches
=================================================================
This examples demonstrates how to decompose a greyscale image
into a number of characteristic rectangular patches.

Lateral inhibition is used to promote localized activations
while circular reconstructions ensures to avoid boundary effects
that would be visible in the atoms and the reconstruction.
"""
import logging

import numpy as np
import matplotlib.pyplot as plt

from tnmf.TransformInvariantNMF import TransformInvariantNMF
from tnmf.utils.data_loading import racoon_image

print(__doc__)

# Activate logging output to ensure fitting progress is printed to the command line.
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s: %(message)s')

# Load an example image.
img = racoon_image(scale=0.1)

# Construct a TransformInvariantNMF instance with selected parameters for the model.
n_atoms = 16
nmf = TransformInvariantNMF(
    n_atoms=n_atoms,
    atom_shape=(6, 10),
    n_iterations=500,
    reconstruction_mode='circular',
    backend='pytorch',
    verbose=3,
)

# Samples provided to nmf.fit() have to be indexed as V[sample, channel, sample_dim_1 .. sample_dim_n].
# This example uses one single-channel image.
img = img[np.newaxis, np.newaxis, ...]

# Run the fitting, i.e. compute dictionary W and activations H so that img = H*W.
nmf.fit(img, inhibition_strength=0.05)

# Collect results from the TransformInvariantNMF instance.
img_r = nmf.R
img_r_partial = np.array([nmf.R_partial(i_atom) for i_atom in range(n_atoms)])
dictionary = nmf.W
activations = nmf.H

# Create a plot of the original image and the reconstruction.
fig, axes = plt.subplots(ncols=2, squeeze=False, figsize=(8, 4))
for (data, title, kwargs), ax in zip([
        (img[0, 0], 'Original', dict(cmap='Greys')),
        (img_r[0, 0], 'Reconstruction', dict(cmap='Greys')),
        ], axes.flatten()):
    ax.set_title(title)
    ax.imshow(data, **kwargs, vmin=0., vmax=1.)
plt.tight_layout()
plt.show()

# Create a plot of the dictionary atoms W, the respective activations H and partial reconstructions.
for data, title, kwargs in [
        (dictionary[:, 0], 'Dictionary Atoms W', dict(cmap='Greys')),
        (activations[0, :], 'Atom Activations H', dict(cmap='gist_ncar')),
        (img_r_partial[:, 0, 0], 'Partial Reconstructions', dict(cmap='Greys')),
]:
    plot_rows = 4
    fig, axes = plt.subplots(nrows=plot_rows, ncols=n_atoms // plot_rows, squeeze=False, figsize=(8, 8),
                             subplot_kw=dict(xticks=[], yticks=[]))
    fig.suptitle(title)
    for i_atom, ax in enumerate(axes.flatten()):
        ax.set_title(i_atom)
        ax.imshow(data[i_atom], **kwargs)
    plt.tight_layout()
    plt.show()



"""
A module that provides some specializations and utilities for all NumPy based backends
that are using FFT for performing convolutions.
"""
from typing import Tuple, Optional

import numpy as np

from ._NumPyBackend import NumPyBackend


class NumPyFFTBackend(NumPyBackend):
    r"""
    The parent class for all NumPy based backends that are using FFT for performing convolutions.

    They provide the functionality to evaluate the analytic gradients of the factorization model.
    """
    def __init__(
        self,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.fft_params = {}

    def _initialize_matrices(
        self,
        V: np.ndarray,
        atom_shape: Tuple[int, ...],
        n_atoms: int,
        W: Optional[np.ndarray] = None,
    ) -> Tuple[np.ndarray, np.ndarray]:

        # this sets pad_mode and pad_width properly
        w, h = super()._initialize_matrices(V, atom_shape, n_atoms, W)

        # fft details: reconstruction
        lower_idx = np.array(atom_shape) - 1
        self.fft_params['reconstruct'] = {
            'fft_axes': self._shift_dimensions,
            'pad_mode': self._pad_mode,
            'pad_width': self._padding_left,
            'correlate': False,
            'slices': tuple(slice(f, f + s) for f, s in zip(lower_idx, self._sample_shape)),
        }

        # fft details: gradient H computation
        lower_idx = np.zeros_like(self._transform_shape)
        if self._pad_mode is not None:
            lower_idx += np.asarray(self._padding_right)[:, 1]
        self.fft_params['grad_H'] = {
            'fft_axes': self._shift_dimensions,
            'pad_mode': self._pad_mode,
            'pad_width': self._padding_right,
            'correlate': True,
            'slices': tuple(slice(f, f + s) for f, s in zip(lower_idx, self._transform_shape)),
        }

        # fft details: gradient W computation
        lower_idx = np.minimum(np.array(self._sample_shape), np.array(self._transform_shape)) - 1
        self.fft_params['grad_W'] = {
            'fft_axes': self._shift_dimensions,
            'pad_mode': self._pad_mode,
            'pad_width': self._padding_right,
            'correlate': True,
            'slices': tuple(slice(f, f + s) for f, s in zip(lower_idx, atom_shape)),
        }

        return w, h
